// NumaAlloc.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#pragma comment(lib, "psapi.lib")

static const ULONGLONG MEGABYTES = (1024 * 1024);

/* pseudo-const */ DWORD  PageSize;  // written once at startup and never changed

void DumpNumaNodeInfo (PVOID Buffer, SIZE_T Size);

/****************************************************************************
*                                 ReportError
* Inputs:
*       LPCTSTR msg: First part of message
*       DWORD err: Error code
* Result: void
*       
* Effect: 
*       Issues the error message
****************************************************************************/

static void ReportError(LPCTSTR msg, DWORD err)
    {
     LPTSTR errmsg;
     if(::FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
                        NULL, // source
                        err,  // Message ID
                        0,    // Default language ID
                        (LPTSTR)&errmsg, // buffer
                        0,    // size
                        NULL) == 0) // argument list
        { /* No message code found */
         _tprintf(_T("%s: Error %d (0x%08x)\n"), msg, err, err);
         return;
        } /* No message code found */
     else
        { /* has message */
         _tprintf(_T("%s: %s\n"), msg, errmsg);
         LocalFree(errmsg);
         return;
        } /* has message */
    } // ReportError

/****************************************************************************
*                              class BufferInfo
****************************************************************************/

class BufferInfo {
    public:
       BufferInfo() { Buffer = NULL; }
       virtual ~BufferInfo() {
                              if(Buffer != NULL)
                                 ::VirtualFree(Buffer, 0, MEM_RELEASE);
                             }
    PVOID Buffer;
};



/****************************************************************************
*                                   _tmain
* Inputs:
*       int argc: Number of pages to allocate
*       _TCHAR * argv[]: Array of arguments
* Result: int
*       0 if successful
*       1 if error
* Effect: 
*       The program
* Argv:
*       [1] If present, the number of pages to allocate, as a decimal number
*           If argc < 2, defaults to 4096 pages
****************************************************************************/

int _tmain(int argc, _TCHAR* argv[])
   {
    ULONG NumberOfProcessors;
    SIZE_T AllocationSize = 0;

    //
    // Get the number of processors and system page size.
    //

    SYSTEM_INFO SystemInfo;
    GetSystemInfo (&SystemInfo);
    NumberOfProcessors = SystemInfo.dwNumberOfProcessors;
    PageSize = SystemInfo.dwPageSize;

    if (argc > 1)
       {
        AllocationSize = _ttoi(argv[1]) * PageSize;
       }

    if (AllocationSize == 0)
       {
        // If not specified, default to 4096 pages
        AllocationSize = 4096 * PageSize;
       }


    //****************************************************************
    // Get the highest node number.
    //****************************************************************

    ULONG HighestNodeNumber;

    if (!GetNumaHighestNodeNumber (&HighestNodeNumber))
       {
        DWORD err = GetLastError();
        ReportError(_T("GetNumaHighestNodeNumber failed"), err);
        return 1;
       }

    if (HighestNodeNumber == 0)
       {
        _putts (_T("Not a NUMA system - exiting\n"));
        return 1;
       }

    //****************************************************************
    // Allocate array of pointers to memory blocks.
    //****************************************************************

    std::vector<BufferInfo> Buffers;
    Buffers.resize(NumberOfProcessors);

    //****************************************************************
    // For each processor, get its associated NUMA node and allocate some memory from it.
    //****************************************************************

    for (UCHAR i = 0; i < NumberOfProcessors; i++)
       { /* Per-processor allocation */
        UCHAR NodeNumber;

        if (!GetNumaProcessorNode (i, &NodeNumber))
           {
            DWORD err = ::GetLastError();
            ReportError(_T("GetNumaProcessorNode failed"), err);
            return 1;
           }

        _tprintf (_T("----------------------------------------------------------------\n"));
        _tprintf (_T("CPU %u: node %u\n"), (ULONG)i, NodeNumber);

        PCHAR Buffer = (PCHAR)VirtualAllocExNuma(
           GetCurrentProcess(),
           NULL,
           AllocationSize,
           MEM_RESERVE | MEM_COMMIT,
           PAGE_READWRITE,
           NodeNumber
                                                );

        if (Buffer == NULL)
           {
            DWORD err = ::GetLastError();
            _tprintf(_T("Node %u, "), NodeNumber);
            ReportError(_T("VirtualAllocExNuma failed"), err);
            return 1;
           }

        PCHAR BufferEnd = Buffer + AllocationSize - 1;
        SIZE_T NumPages = ((SIZE_T)BufferEnd)/PageSize - ((SIZE_T)Buffer)/PageSize + 1;

        _tprintf (_T("Allocated virtual memory: %p - %p (%6Iu pages), preferred node %u\n"), Buffer, BufferEnd, NumPages, NodeNumber);

        Buffers[i].Buffer = Buffer;

        //****************************************************************
        // At this point, virtual pages are allocated but no valid physical
        // pages are associated with them yet.
        //
        // The FillMemory call below will touch every page in the buffer, faulting
        // them into our working set. When this happens physical pages will be allocated
        // from the preferred node we specified in VirtualAllocExNuma, or any node
        // if the preferred one is out of pages.
        //****************************************************************

        FillMemory (Buffers[i].Buffer, AllocationSize, 'x');

        //****************************************************************
        // Check the actual node number for the physical pages that are still valid
        // (if system is low on physical memory, some pages could have been trimmed already).
        //****************************************************************

        DumpNumaNodeInfo (Buffers[i].Buffer, AllocationSize);
       } /* Per-processor allocation */

    return 0;
   } // _tmain

/****************************************************************************
*                                 DumpRegion
* Inputs:
*       PVOID StartPtr: Starting address of the region
*       PVOID EndPtr: Ending address of the region
*       BOOL Valid: TRUE if the region is valid, FALSE otherwise
*       DWORD Node: NUMA node this applies to
* Result: void
*       
* Effect: 
*       Prints out the message
****************************************************************************/

void DumpRegion (PVOID StartPtr, PVOID EndPtr, BOOL Valid, DWORD Node)
   {
    DWORD_PTR StartPage = ((DWORD_PTR)StartPtr)/PageSize;
    DWORD_PTR EndPage   = ((DWORD_PTR)EndPtr)/PageSize;
    DWORD_PTR NumPages  = (EndPage - StartPage) + 1;

    if (Valid)
       {
        _tprintf (_T("%p - %p (%6Iu pages): node %u\n"), StartPtr, EndPtr, NumPages, Node);
       }
    else
       {
        _tprintf (_T("%p - %p (%6Iu pages): no valid pages\n"), StartPtr, EndPtr, NumPages);
       }
   }

/****************************************************************************
*                              DumpNumaNodeInfo
* Inputs:
*       PVOID Buffer: A pointer to a block of memory
*       SIZE_T Size: The size of the block of memory, in bytes
* Result: void
*       
* Effect: 
*       Displays how the memory is spread across the NUMA nodes
****************************************************************************/

void DumpNumaNodeInfo (PVOID Buffer, SIZE_T Size)
   {
    DWORD_PTR StartPage = ((DWORD_PTR)Buffer)/PageSize;
    DWORD_PTR EndPage   = ((DWORD_PTR)Buffer + Size - 1)/PageSize;
    DWORD_PTR NumPages  = (EndPage - StartPage) + 1;

    PCHAR StartPtr = (PCHAR)(PageSize*StartPage);

    _tprintf(_T("Checking NUMA node for %d pages:\n"), NumPages);

    std::vector<PSAPI_WORKING_SET_EX_INFORMATION> WsInfo;
    WsInfo.resize(NumPages);

    for (DWORD_PTR i = 0; i < NumPages; i++)
       {
        WsInfo[i].VirtualAddress = StartPtr + (i * PageSize);
       }

    BOOL bResult = QueryWorkingSetEx(
                                     GetCurrentProcess(),
                                     &WsInfo[0],
                                     NumPages * sizeof(PSAPI_WORKING_SET_EX_INFORMATION)
                                    );

    if (!bResult)
       {
        DWORD err = ::GetLastError();
        ReportError (_T("QueryWorkingSetEx failed"), err);
        return;
       }

    PCHAR RegionStart;
    BOOL  RegionIsValid;
    DWORD RegionNode;

    for (DWORD_PTR i = 0; i < NumPages; i++)
       { /* Show runs */
        PCHAR Address = (PCHAR)WsInfo[i].VirtualAddress;
        BOOL  IsValid = WsInfo[i].VirtualAttributes.Valid;
        DWORD Node    = WsInfo[i].VirtualAttributes.Node;

        if (i == 0)
           { /* first run */
            RegionStart   = Address;
            RegionIsValid = IsValid;
            RegionNode    = Node;
           } /* first run */

        if (IsValid != RegionIsValid || Node != RegionNode)
           { /* run change */
            DumpRegion (RegionStart, Address - 1, RegionIsValid, RegionNode);

            RegionStart   = Address;
            RegionIsValid = IsValid;
            RegionNode    = Node;
           } /* run change */

        if (i == (NumPages - 1))
           { /* end of runs */
            DumpRegion (RegionStart, Address + PageSize - 1, IsValid, Node);
           } /* end of runs */
       } /* Show runs */
   } // DumpNumaNodeInfo


